/*
 * completion.c
 *
 * Copyright (C) 2006  Insigme Co., Ltd
 *
 * This software has been developed while working on the Linux Unified Kernel
 * project (http://www.longene.org) in the Insigma Research Institute,  
 * which is a subdivision of Insigma Co., Ltd (http://www.insigma.com.cn).
 * 
 * The project is sponsored by Insigma Co., Ltd.
 *
 * The authors can be reached at linux@insigma.com.cn.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of  the GNU General  Public License as published by the
 * Free Software Foundation; either version 2 of the  License, or (at your
 * option) any later version.
 *
 * Revision History:
 *   Dec 2008 - Created.
 */
 
/*
 * completion.c:
 * Refered to Wine code
 */
#include <linux/mm.h>
#include <linux/syscalls.h>

#include "wineserver/server.h"
#include "wineserver/request.h"
#include "wineserver/uk_lib.h"
#include "io.h"
#include "object.h"
#include "unistr.h"
#include "event.h"
#include "section.h"
#include "handle.h"
#include "objwait.h"
#include "winuser.h"
#include "wineserver/winerror.h"
#include "thread.h"
#include "wineserver/file.h"
#include "wineserver/user.h"
#include "wineserver/winerror.h"

#ifdef CONFIG_UNIFIED_KERNEL
struct uk_completion
{
    struct object  obj;
    struct list_head    queue;
    unsigned int   depth;
};

static void completion_dump(struct object*, int);
static struct object_type *completion_get_type(struct object *obj);
static void completion_destroy(struct object *);

static const struct object_ops completion_ops =
{
    sizeof(struct uk_completion), /* size */
    completion_dump,           /* dump */
    completion_get_type,       /* get_type */
    no_get_fd,                 /* get_fd */
    no_map_access,             /* map_access */
    no_lookup_name,            /* lookup_name */
    no_open_file,              /* open_file */
    no_close_handle,           /* close_handle */
    completion_destroy         /* destroy */
};

struct comp_msg
{
    struct   list_head queue_entry;
    unsigned long ckey;
    unsigned long cvalue;
    unsigned long information;
    unsigned int  status;
};

static void completion_destroy(struct object *obj)
{
    struct uk_completion *completion = (struct uk_completion *) obj;
    struct comp_msg *tmp, *next;

    LIST_FOR_EACH_ENTRY_SAFE(tmp, next, &completion->queue, struct comp_msg, queue_entry) {
        free(tmp);
    }
}

static void completion_dump(struct object *obj, int verbose)
{
#if 0
    struct uk_completion *completion = (struct uk_completion *) obj;

    assert(obj->ops == &completion_ops);
    fprintf(stderr, "Completion ");
    dump_object_name(&completion->obj);
    fprintf(stderr, " (%u packets pending)\n", completion->depth);
#endif
}

extern struct object_type *get_object_type(const struct unicode_str *name);

static struct object_type *completion_get_type(struct object *obj)
{
    static const WCHAR name[] = {'C','o','m','p','l','e','t','i','o','n'};
    static const struct unicode_str str = { name, sizeof(name) };
    return get_object_type(&str);
}

static struct uk_completion *create_completion(struct directory *root, const struct unicode_str *name, unsigned int attr, unsigned int concurrent)
{
    struct uk_completion *completion;

    if ((completion = create_named_object_dir(root, name, attr, &completion_ops))) {
        if (get_error() != STATUS_OBJECT_NAME_EXISTS) {
            INIT_LIST_HEAD(&completion->queue);
            completion->depth = 0;
        }
    }

    return completion;
}

struct uk_completion *get_completion_obj(struct w32process *process, obj_handle_t handle, unsigned int access)
{
    return (struct uk_completion *)get_wine_handle_obj(process, handle, access, &completion_ops);
}

extern void uk_wake_up(struct object *obj, int max);

void add_completion(struct uk_completion *completion, unsigned long ckey, unsigned long cvalue, unsigned int status, unsigned long information)
{
    struct comp_msg *msg = mem_alloc(sizeof(*msg));

    if (!msg)
        return;

    msg->ckey = ckey;
    msg->cvalue = cvalue;
    msg->status = status;
    msg->information = information;
    list_add_tail(&completion->queue, &msg->queue_entry);
    completion->depth++;
    uk_wake_up(&completion->obj, 1);
}

/* create a completion */
DECL_HANDLER(create_completion)
{
    struct uk_completion *completion;
    struct unicode_str name;
    struct directory *root = NULL;

    reply->handle = 0;

    get_req_unicode_str(&name);

    if ((completion = create_completion(req->rootdir, &name, req->attributes, req->concurrent)) != NULL) {
        reply->handle = alloc_handle(get_current_w32process(), completion, req->access, req->attributes);
        release_object(completion);
    }

    if (root) release_object(root);
}

/* open a completion */
DECL_HANDLER(open_completion)
{
    struct uk_completion *completion;
    struct unicode_str name;
    struct directory *root = NULL;

    reply->handle = 0;

    get_req_unicode_str(&name);

    if ((completion = open_object_dir(req->rootdir, &name, req->attributes, &completion_ops)) != NULL) {
        reply->handle = alloc_handle(get_current_w32process(), completion, req->access, req->attributes);
        release_object(completion);
    }

    if (root)
        release_object(root);
}

#define IO_COMPLETION_QUERY_STATE  0x0001
#define IO_COMPLETION_MODIFY_STATE 0x0002
#define IO_COMPLETION_ALL_ACCESS   (STANDARD_RIGHTS_REQUIRED|SYNCHRONIZE|0x3)

/* add completion to completion port */
DECL_HANDLER(add_completion)
{
    struct uk_completion* completion = get_completion_obj(get_current_w32process(), req->handle, IO_COMPLETION_MODIFY_STATE);

    if (!completion)
        return;

    add_completion(completion, req->ckey, req->cvalue, req->status, req->information);

    release_object(completion);
}

/* get completion from completion port */
DECL_HANDLER(remove_completion)
{
    struct uk_completion* completion = get_completion_obj(get_current_w32process(), req->handle, IO_COMPLETION_MODIFY_STATE);
    struct list_head *entry;
    struct comp_msg *msg;

    if (!completion)
        return;

    entry = list_head(&completion->queue);
    if (!entry)
        set_error(STATUS_PENDING);
    else {
        list_del(entry);
        completion->depth--;
        msg = LIST_ENTRY(entry, struct comp_msg, queue_entry);
        reply->ckey = msg->ckey;
        reply->cvalue = msg->cvalue;
        reply->status = msg->status;
        reply->information = msg->information;
        free(msg);
    }

    release_object(completion);
}

/* get queue depth for completion port */
DECL_HANDLER(query_completion)
{
    struct uk_completion* completion = get_completion_obj(get_current_w32process(), req->handle, IO_COMPLETION_QUERY_STATE);

    if (!completion)
        return;

    reply->depth = completion->depth;

    release_object(completion);
}
#endif /* CONFIG_UNIFIED_KERNEL */
